import os
import re
from typing import List, Optional, Tuple, Type

from karadoc.common.conf import get_model_folder_location
from karadoc.common.job_core.job_base import JobBase
from karadoc.common.model import file_index


def _model_location():
    return get_model_folder_location()


def get_data_location(table_name: str) -> str:
    (schema_name, table_name, _) = parse_table_name(table_name)
    return "data/hive/warehouse/%s.db/%s" % (schema_name, table_name)


def get_folder_location(table_name: str, job_type: Type[JobBase]) -> Optional[str]:
    (schema_name, table_name, _) = parse_table_name(table_name)

    path = file_index.get_action_file(schema_name, table_name, job_type)
    if path:
        return os.path.split(path)[0]
    else:
        return None


def parse_partition_name(partitions: Optional[str]) -> Optional[List[Tuple[str, str]]]:
    if partitions:
        res = []
        for p in partitions.split("/"):
            [k, v] = p.split("=")
            res.append((k, v))
        return res
    else:
        return None


def remove_partition_from_name(full_table_name: str) -> str:
    schema, table, _ = parse_table_name(full_table_name)
    return "%s.%s" % (schema, table)


def parse_table_name(table_name: str) -> Tuple[str, str, str]:
    t = """[_A-Za-z0-9]+"""
    table_regex: re.Pattern[str] = re.compile(rf"""\A({t})[.]({t})(?:/(.*))?\Z""")
    matcher = table_regex.match(table_name)
    if matcher:
        [schema_name, table_name, partitions] = matcher.groups()
        return schema_name, table_name, parse_partition_name(partitions)
    else:
        raise ValueError("Incorrect table name : %s" % table_name)


def table_exists(table_name: str, job_type: Type[JobBase]) -> bool:
    folder_location = get_folder_location(table_name, job_type)
    if folder_location:
        return os.path.isdir(folder_location)
    else:
        return False


def populate_exists(full_table_name: str) -> bool:
    """
    :return: True if the given table has a {spark.batch}.py
    """
    from karadoc.common.run.spark_batch_job import SparkBatchJob

    return __action_file_exists(full_table_name, SparkBatchJob)


def quality_check_exists(full_table_name: str) -> bool:
    """
    :return: True if the given table has a {spark.quality_check}.py
    """
    from karadoc.common.quality.quality_check_job import QualityCheckJob

    return __action_file_exists(full_table_name, QualityCheckJob)


def stream_file_exists(full_table_name: str) -> bool:
    """
    :return: True if the given table has a {stream.batch}.py
    """
    from karadoc.common.stream.spark_stream_job import SparkStreamJob

    return __action_file_exists(full_table_name, SparkStreamJob)


def __action_file_exists(full_table_name: str, job_type: Type) -> bool:
    """
    :return: True if the given table has an ACTION_FILE.py
    """
    (schema_name, table_name, _) = parse_table_name(full_table_name)
    path = file_index.get_action_file(schema_name, table_name, job_type)
    if path is None:
        return False
    else:
        return os.path.isfile(path)



from argparse import ArgumentParser
from typing import List

from termcolor import colored

from karadoc.common.commands.command import Command
from karadoc.common.commands.return_code import ReturnCode
from karadoc.common.observability.conf import setup_observability
from karadoc.common.validations import ValidationResult, ValidationSeverity


def count_results_with_severity(validation_results: List[ValidationResult], severity: ValidationSeverity) -> int:
    return len([res for res in validation_results if res.severity == severity])


def inspect_results(results: List[ValidationResult]):
    """Function used for unit tests. Mock this function to be able to inspect the produced results
    during the execution of the command."""
    pass


class ValidateAirflowCommand(Command):
    description = "perform various checks on the airflow dags"

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        parser.add_argument(
            "--coverage",
            dest="coverage",
            default=False,
            action="store_true",
            help="Activate this option to generate code coverage report on the files that where tested.",
        )
        parser.add_argument(
            "--max-warnings",
            dest="max_warnings",
            metavar="N",
            type=int,
            help="Returns an error code if more than N warning are found.",
        )

    @staticmethod
    def do_command(args) -> ReturnCode:
        return_code = ReturnCode.Success
        cov = None
        if args.coverage:
            import coverage

            cov = coverage.Coverage()
            cov.start()

        from karadoc.airflow.dag_graph import build_dag_bag
        from karadoc.airflow.validate import validate_dag_bag
        from karadoc.common.model import table_index

        table_index = table_index.build_table_index()
        dag_bag = build_dag_bag()

        # We force the reset of the logging conf after loading the DAG bag because Airflow overrode our configuration
        setup_observability(reset=True)

        validation_results = validate_dag_bag(dag_bag, table_index)

        if len(validation_results) > 0:
            inspect_results(validation_results)

        for result in validation_results:
            print(result)

        num_errors = count_results_with_severity(validation_results, ValidationSeverity.Error)
        if num_errors > 0:
            print(colored(f"Found {num_errors} errors", "red"))
            return_code = ReturnCode.Error

        num_warnings = count_results_with_severity(validation_results, ValidationSeverity.Warning)
        if args.max_warnings is not None and num_warnings > args.max_warnings:
            print(
                colored(
                    f"Found {num_warnings} warnings, which is greater than option --max-warnings {args.max_warnings}",
                    "red",
                )
            )
            return_code = ReturnCode.Error

        if cov is not None:
            cov.stop()
            cov.save()

        return return_code
